# -*- coding: utf-8 -*-
from django import forms
from django.contrib.admin.widgets import AdminFileWidget

from tcms.models import Page, Path, Value, WIP
from tcms.tpl import mkbasename
from tcms.fields import PathWidget, AdminTexareaField, AdminCharField


class PageForm(forms.ModelForm):
    """Edit page form"""
    path = forms.CharField(label='URL', widget=PathWidget())
    description = AdminTexareaField(required=False)
    search_image = forms.ImageField(label='Image', widget=AdminFileWidget, required=False)
    search_text = AdminTexareaField(label='Text', required=False)
    meta_title = AdminCharField(label='Title', required=False)
    meta_keywords = AdminCharField(label='Keywords', required=False)
    meta_description = AdminTexareaField(label='Description', required=False)

    def clean_path(self):
        """Path cleaning"""
        return _clean_path(self)

    def page_fields(self):
        """Returns page concerning fields, handly for template"""
        for field in self:
            name = field.html_name
            if not (name.startswith('meta_') or name.startswith('search_')):
                yield field

    def metadata_fields(self):
        """Returns page metadata fields, handly for template"""
        for field in self:
            if field.html_name.startswith('meta_'):
                yield field

    def search_fields(self):
        """Returns page search fields, handly for template"""
        for field in self:
            if field.html_name.startswith('search_'):
                yield field

    class Meta:
        model = Page
        exclude = ('template', 'state', 'updated')


class CopyPageForm(forms.ModelForm):
    """Copy page form"""
    path = forms.CharField(label='URL', widget=PathWidget())

    def clean_path(self):
        """Path cleaning"""
        path = _clean_path(self)
        # validate that there's no another WIP page with same path
        if path.pages.filter(state=WIP).count() > 0:
            raise forms.ValidationError(
                    'Another "Work in progress" page exist for given URL')
        return path

    def save(self, orig):
        """Copy and return new page"""
        return orig.copy(self.cleaned_data['path'])

    class Meta:
        model = Page
        fields = ('path',)


class ImportForm(forms.Form):
    """Page importing form"""
    file = forms.FileField(label='Exported file', required=True)

    def clean_file(self):
        """File cleaning, validates it using Page.valid_xml()"""
        file = self.cleaned_data['file']
        try:
            Page.valid_xml(file.read())
        except forms.ValidationError, e:
            raise e
        else:
            file.seek(0) # seek file because will be used again
            return file

    def save(self):
        return Page.from_xml(self.cleaned_data['file'].read())


class DynamicForm(forms.Form):
    """Dynamic form. Will create a form from a set of (name, type) pairs."""
    basename = forms.CharField(widget=forms.HiddenInput(), required=True)

    def __init__(self, names, *args, **kwargs):
        """Init method, will add fields according to name and type."""
        super(DynamicForm, self).__init__(*args, **kwargs)
        self.names = names
        for name, type in names:
            self.fields[name] = type.as_field(label=name.replace('_', ' '),
                                              required=False)

    def save(self, page):
        """Save method. Will save each section value.
        Will return saved values"""
        result = []
        basename = self.cleaned_data['basename']

        for name, type in self.names:
            try: # use type save handler
                value = type.to_database(self, name)
            except ValueError: # ignore error
                pass
            else:
                vname = mkbasename(basename, name)
                try: # try to update a Value instance
                    obj = page.values.get(name=vname)
                    # if types are different, like it was changed in page
                    # definition, needs to be overriden to avoid duplications
                    # because of uniqueness definition on Value model
                    if obj.type != type.name():
                        obj.type = type.name()
                    obj.value = value
                    obj.save()
                except Value.DoesNotExist: # save new value
                    obj = page.values.create(name=vname, type=type.name(),
                                             value=value)
                result.append(obj)
        return result


def _clean_path(form, field='path'):
    """Retrieve Path object or raise ValidationError if not exists"""
    try:
        path = Path.objects.get(pk=form.cleaned_data[field])
    except Path.DoesNotExist:
        raise forms.ValidationError('Invalid %s' % field)
    return path



from django.contrib import sitemaps
from django.core import urlresolvers
from models import Page, LIVE

class TCMSSitemap(sitemaps.Sitemap):
    """Return the sitemap items from CMS"""
    def __init__(self, patterns=None, priority=None, changefreq=None):
        """
        patterns is list of url patterns to find CMS pages for. Should have "name" param defined
        when patterns are ommited all live cms pages are used
        
        priority and changefreq allows to define eponymous params in resulting sitemap
        """
        
        if hasattr(patterns, "__iter__"):
            self.patterns = [pattern.name for pattern in patterns 
                if hasattr(pattern, 'name') and pattern.name]
        else:
            self.patterns = None

        self.priority = priority
        self.changefreq = changefreq

    def items(self):
        pages = Page.objects.select_related('path').filter(state = LIVE)

        if self.patterns != None:
            urls = map(urlresolvers.reverse, self.patterns)
            urls = [url + '/' if not url.endswith('/') else url for url in urls]
            
            pages = list(pages.filter(path__path__in = urls))
            
        return pages
        

    def changefreq(self, obj):
        return self.changefreq

    def priority(self, obj):
        return self.priority

    def lastmod(self, obj):
        return obj.updated

    def location(self, obj):
        return str(obj.path)


# -*- coding: utf-8 -*-
import re
import base64
from os.path import split
from cStringIO import StringIO

from django.conf import settings
from django.core.cache import cache
from django.core.files.uploadedfile import InMemoryUploadedFile
from django.db.models.fields.files import ImageFieldFile
from django.utils.datastructures import DotExpandedDict


CACHE_NAME = getattr(settings, 'TCMS_CACHE_NAME', 'tcms')

# Capital letters regex
CAPLETTERS = re.compile('([A-Z])')

def human_title(value):
    """Converts camel-case word to human readeable format."""
    return ' '.join(word.title()
                        for word in CAPLETTERS.sub(r' \1', value)\
                                              .replace('_', ' ')\
                                              .split())


def update_cache():
    """Updates CMS pages cache, cache structure is a dictioray which key
    is a tuple of page path, locale and state, and valud is page id."""
    from tcms.models import Page, WIP, LIVE

    pages = Page.objects.filter(state__in=[WIP, LIVE]).select_related('path')

    if getattr(settings, 'TCMS_LOCALIZED', False):
        values = dict(((page.path.path, page.path.locale, page.state), page.id)
                            for page in pages)
    else:
        values = dict(((page.path.path, page.state), page.id)
                            for page in pages)
    cache.set(CACHE_NAME, values)
    return values


def locale_tags(locale):
    """ Splits a locale/language tag into a subtags sequence
    >>> locale_tags('en')
    ('en', '')
    >>> locale_tags('en-gb')
    ('en-gb', 'en', '')
    >>> locale_tags('')
    ('',)
    """
    if not locale:
        return ('',)
    return (locale,) + locale_tags(locale.rpartition('-')[0])


def id_from_cache(paths, locale=None):
    """Load page id from cache if present. Will try locale and sub locales
    withing LIVE and WIP states."""
    from tcms.models import WIP, LIVE

    values = cache.get(CACHE_NAME)
    if values is None: # load cache if not entry
        values = update_cache()

    if not isinstance(paths, (list, tuple)):
        paths = [paths]

    locales = locale_tags(locale)
    for path in paths: # test each path for possible locales
        path = normalize_path(path)
        if getattr(settings, 'TCMS_LOCALIZED', False):
            for loc in locales:
                if (path, loc, LIVE) in values:
                    return values[(path, loc, LIVE)]
                elif (path, loc, WIP) in values:
                    return values[(path, loc, WIP)]
        else:
            if (path, LIVE) in values:
                return values[(path, LIVE)]
            elif (path, WIP) in values:
                return values[(path, WIP)]


def save_b64_image(value, name, model_field, save=False):
    """
    Saves an base64 encoded string as a file using @model_field
    (which must be an django.db.models.fields.files.ImageFieldFile
    instance) save method. @name is file name to save to, but might
    gain suffixes on coincidence. Model instance will be saved by
    django if @save flag is True. Returns saved file name or none
    if @value and @name are not valid.
    """
    if value and name and isinstance(model_field, ImageFieldFile):
        value = base64.decodestring(value)
        sio = StringIO()
        sio.write(value)
        sio.seek(0)
        data = InMemoryUploadedFile(file=sio, name=name, field_name=None,
                                    content_type=None, size=len(value),
                                    charset=None)
        model_field.save(name, data, save=save)
        return model_field.name


def image_to_b64(model_field):
    """Encodes image value in base64 encoding. Returns tuple with
    file name and encoded content, None in case of error.
    @model_field must be an django.db.models.fields.files.ImageFieldFile
    instance.
    """
    if isinstance(model_field, ImageFieldFile):
        try:
            return (split(model_field.name)[-1],
                    base64.encodestring(model_field.read(model_field.size)))
        except: # ignore exceptions
            pass


def normalize_path(path):
    """Normalize a path to be used on CMS. Basically, appends a trailing slahs"""
    path = path.strip('/')
    path = ('/' + path + '/') if path else '/'
    return path


def dotted_dict_to_choices(dict_, sort=True):
    """Coverts a dotted dictionary (see DotExpandedDict class for details)
    to a recursive choices format which will be used to form <optgroup> tags
    when used with django fields using choices attribute."""
    def _flat(name, values, first=False):
        """Flatten structure, recursive."""
        if isinstance(values, dict):
            choices = [_flat(name + '.' + k, v) for k, v in values.iteritems()]
        elif not first:
            return (name, values.NAME)
        else:
            choices = [(name, values.NAME)]
        if sort:
            choices.sort(key=lambda x: x[0])
        return (name.title(), choices)

    choices = [_flat(k, v, True)
                    for k, v in DotExpandedDict(dict_).iteritems()]
    if sort:
        choices.sort(key=lambda (x, y): x if isinstance(y, list) else y)
    return choices



# -*- coding: utf-8 -*-
from django.conf import settings
from django.http import HttpResponse, HttpResponseBadRequest, \
                        HttpResponseRedirect, Http404
from django.core.urlresolvers import reverse
from django.shortcuts import get_object_or_404, render_to_response
from django.utils import simplejson, html
from django.utils.encoding import force_unicode
from django.template.context import RequestContext
from django.template.defaultfilters import slugify
from django.contrib import messages
from django.contrib.admin.models import LogEntry, CHANGE, ADDITION, DELETION
from django.contrib.admin.views.main import IS_POPUP_VAR, SEARCH_VAR
from django.contrib.contenttypes.models import ContentType
from django.contrib.auth.decorators import user_passes_test

from tcms.models import Page, TYPES_MAP
from tcms.data_types import RawIdType
from tcms.forms import PageForm, CopyPageForm, ImportForm
from tcms.utils import update_cache
from tcms.exceptions import TemplateFormValidationError


class TextareaAjaxResponse(HttpResponse):
    """Response that wraps content in <textarea></textarea> for usage with
    jquery.forms.js plugin.
    """
    def __init__(self, data='', *args, **kwargs):
        super(TextareaAjaxResponse, self).__init__(*args, **kwargs)
        content = simplejson.dumps(data)
        self.content = '<textarea>%s</textarea>' % html.escape(content)


def permission_required(*perms):
    """Permissions checking decorator"""
    return user_passes_test(lambda u: u.has_perms(perms))


@permission_required('change_page')
def edit(request, page_id):
    """Page edition view"""
    page = get_object_or_404(Page, pk=page_id)
    page.load()

    if request.method == 'POST':
        form = PageForm(instance=page, data=request.POST, files=request.FILES)
        if form.is_valid():
            form.save()
            log(request, page, CHANGE, 'Page edited')
            messages.info(request, 'Page saved correctly')
            return HttpResponseRedirect(_edit_url(page_id))
    else:
        form = PageForm(instance=page)
    state = page.get_state_display()
    return _render('cms/edit.html', locals(), request)



@permission_required('add_value', 'change_value')
def edit_section(request, page_id, section):
    """Page section edition view"""
    page = get_object_or_404(Page, pk=page_id)

    if request.method == 'POST':
        basename = request.POST.get('basename')
        if not basename:
            raise AttributeError('Invalid basename')

        try:
            page.tpl.save(page, basename, data=request.POST,
                          files=request.FILES)
        except TemplateFormValidationError, e:
            if request.POST.get('is_ajax'):
                errors = [{'field': name, 'messages': list(errors)}
                                    for name, errors in e.errors.iteritems()]
                return TextareaAjaxResponse({'status': 'error',
                                             'errors': errors})
        else:
            log(request, page, CHANGE, 'Content for section %s updated (%s)' % \
                                                (section, basename))
            messages.info(request, 'Section data saved correctly')
            if request.POST.get('is_ajax'):
                return TextareaAjaxResponse({'status': 'ok'})
            else:
                url = _edit_section_url(page_id, section)
                url_hash = basename.replace('/', '-')
                return HttpResponseRedirect(url + '#' + url_hash)
    page.load(section)
    section_name = section  # used in template
    section = page.tpl[section_name]
    state = page.get_state_display()
    TCMS_CKEDITOR_BASE_URL = getattr(settings, 'TCMS_CKEDITOR_BASE_URL', '')
    if TCMS_CKEDITOR_BASE_URL and not TCMS_CKEDITOR_BASE_URL.endswith('/'):
        TCMS_CKEDITOR_BASE_URL += '/'
    return _render('cms/edit_section.html', locals(), request)


@permission_required('delete_value')
def clear(request, page_id, section):
    """Page section data clear. Clears some section values set."""
    if request.method == 'POST':
        page = get_object_or_404(Page, pk=page_id)
        basename = request.POST['basename']
        page.values.filter(name__startswith=basename).delete()
        log(request, page, CHANGE, 'Content for section %s cleared' % section)
        messages.info(request, 'Content for section %s cleared' % section)
    return HttpResponseRedirect(_edit_section_url(page_id, section))


def rawid(request):
    """Display a list to pick a raw id for an object of @type type, like
    django admin raw id feature"""
    Type = TYPES_MAP.get(request.GET.get('type'))
    if not issubclass(Type, RawIdType):
        return HttpResponseBadRequest('Wrong type "%s"' % Type)
    ctx = Type.rawid_context(request)
    ctx['type'] = Type.name()
    ctx['IS_POPUP_VAR'] = IS_POPUP_VAR
    ctx['is_popup'] = request.GET.get(IS_POPUP_VAR, False)
    ctx[SEARCH_VAR] = request.GET.get(SEARCH_VAR, '')
    return _render('cms/rawid.html', ctx, request)


@permission_required('add_page', 'add_path')
def copy(request, page_id):
    """Page copy view. Copies a page under a different URL/Locale"""
    page = get_object_or_404(Page, pk=page_id)

    if request.method == 'POST':
        form = CopyPageForm(request.POST)
        if form.is_valid(): # save page and duplicate values
            dup = form.save(page)
            log(request, page, ADDITION, 'Page copied to object %s (%s)' % \
                                                (dup.id, dup.path.path))
            log(request, dup, ADDITION, 'Page copied from %s (%s)' % \
                                                (page.id, page.path.path))
            messages.info(request, 'Page copied correctly')
            return HttpResponseRedirect(_edit_url(dup.id))
    else:
        form = CopyPageForm()
    return _render('cms/copy.html', locals(), request)


@permission_required('change_page', 'add_rendered', 'change_rendered')
def publish(request, page_id):
    """Page publishing view"""
    page = get_object_or_404(Page, pk=page_id)
    page.publish()
    update_cache()
    log(request, page, CHANGE, 'Page published')
    messages.info(request, 'Page %s published' % page)
    return HttpResponseRedirect(request.GET.get('next') or _edit_url(page_id))


@permission_required('change_page')
def unpublish(request, page_id):
    """Page unpublishing view"""
    page = get_object_or_404(Page, pk=page_id)
    page.unpublish()
    update_cache()
    log(request, page, CHANGE, 'Page unpublished')
    messages.info(request, 'Page %s unpublished' % page)
    return HttpResponseRedirect(request.GET.get('next') or _edit_url(page_id))


@permission_required('add_rendered', 'change_rendered')
def refresh(request, page_id):
    """Page rendered data refreshing view"""
    page = get_object_or_404(Page, pk=page_id)
    page.refresh()
    log(request, page, CHANGE, 'Content refreshed')
    messages.info(request, 'Content for page %s refreshed' % page)
    return HttpResponseRedirect(request.GET.get('next') or _edit_url(page_id))


@permission_required('delete_page', 'delete_value', 'delete_rendered')
def delete(request, page_id):
    """Page deletion view"""
    page = get_object_or_404(Page, pk=page_id)
    try:
        page.delete()
    except TypeError, e:
        messages.info(request, e.message)
        return HttpResponseRedirect(_edit_url(page_id))
    else:
        update_cache()
        log(request, page, DELETION, 'Page deleted')
        messages.info(request, 'Page %s deleted' % page)
        return HttpResponseRedirect(reverse('admin:tcms_page_changelist'))


def export(request, page_id):
    """Page exporting view"""
    page = get_object_or_404(Page.objects.select_related('path'), pk=page_id)
    response = HttpResponse(mimetype='application/x-download')
    page.to_xml(response)
    # homepage path is sluged as '', then we renamed it as 'homepage'
    name = slugify(page.path.path).replace('-', '_') or 'homepage'
    response['Content-Disposition'] = 'attachment; filename=%s.xml' % name
    log(request, page, CHANGE, 'Page exported')
    return response


@permission_required('add_path', 'add_page', 'add_value')
def import_page(request):
    """Page importing view"""
    if request.method == 'POST':
        form = ImportForm(data=request.POST, files=request.FILES)
        if form.is_valid():
            try:
                page = form.save()
            except Exception, e:
                messages.info(request, 'Impossible to import: %s' % e)
            else:
                update_cache()
                log(request, page, ADDITION, 'Page imported')
                messages.info(request, 'Page %s imported' % page)
                return HttpResponseRedirect(_edit_url(page.id))
    else:
        form = ImportForm()
    return _render('cms/import.html', {'form': form}, request)


def log(request, obj, level, message=''):
    """Log events for user notification"""
    ct = ContentType.objects.get_for_model(obj.__class__).pk
    try:
        object_repr = force_unicode(obj)
    except:
        object_repr = ''
    LogEntry.objects.log_action(user_id=request.user.id, content_type_id=ct,
                                object_id=obj.pk, object_repr=object_repr,
                                action_flag=level, change_message=message)


def render_to_template(request, context=None):
    """Renders a page directly to it's default template defined, raises
    404 if current request is not a valid CMS page or if the related
    template doesn't define a template for it."""
    ctx = RequestContext(request)
    page = ctx.get('cms')
    if page is None or page.tpl.TEMPLATE is None:
        raise Http404('Not found')
    return render_to_response(page.tpl.TEMPLATE, context or {}, ctx)


def _edit_url(page_id):
    """Return edition url for @page_id"""
    return reverse('admin:cms_edit', args=(page_id,))

def _edit_section_url(page_id, section):
    """Return section edition url for @page_id/@section"""
    return _edit_url(page_id) + '%s/' % section

def _render(tpl, ctx, req):
    """Renders to response loading request context"""
    return render_to_response(tpl, ctx, RequestContext(req))
